# TigerStyle

Adopted from the original [Tiger Style](https://github.com/tigerbeetle/tigerbeetle/blob/main/docs/TIGER_STYLE.md)
for **zig**, created by tigerbeetle.

## The Essence of Our Go Style

> “There are three things extremely hard: steel, a diamond, and to know one's self.” — Benjamin Franklin

Our Go style is not merely about formatting code—it’s our design philosophy. It blends engineering excellence with artful clarity to ensure that our systems are safe, high-performance, and enjoyable to develop and maintain.

---

## Why Have Style?

In our view, design is as much about how the system works as what it looks like. Our style makes each decision explicit, ensuring that safety, performance, and developer experience remain paramount.

- **Safety** first,
- **Performance** next,
- and then **developer experience**.

Good style does more than improve readability; it enforces the design decisions that prevent technical debt from creeping in.

---

## On Simplicity, Elegance, and Technical Debt

- **Simplicity** is our ultimate goal—but it is never a free pass. Achieving simplicity is the result of thoughtful design and disciplined iteration.
- A “zero technical debt” policy applies: we design and implement things right the first time. This saves each of us from the cost of firefighting production issues later.

---

## Safety

Even in Go, safety is non‐negotiable. We frequently embed explicit checks and assertions to ensure our code behaves as expected. Consider this sample assertion helper:

```go
// Assert panics with a message if the condition is not met.
// Not idiomatic Go, but used here to enforce critical invariants.
func Assert(condition bool, msg string) {
	if !condition {
		panic(fmt.Sprintf("Assertion failed: %s", msg))
	}
}
```

Key safety guidelines:

- Use **simple, explicit control flow** (avoid unnecessary recursion and hidden complex abstractions).
- **Limit loops and queues**: every loop should have an explicit upper bound to avoid infinite cycles.
- Favor fixed-size types (e.g., use `uint32` or `int32` where appropriate) instead of architecture-dependent sizes.
- Use assertions liberally—both to check expected states _before_ and _after_ important operations. Aim for at least two assertions per function.
- All memory usage should be planned up-front. Although Go provides garbage collection, design your modules with a fixed resource strategy in mind for predictability.
- Keep function bodies short. In Go, we suggest striving for functions to be easily digestible (roughly keeping functions around 70 lines or less) by breaking out helper functions where it makes sense.

---

## Performance

Performance starts in the design phase—even before profiling code.

- **Back-of-the-envelope sketches:** Before implementation, roughly gauge resource (network, disk, memory, CPU) usage.
- Identify the slowest resource (network → disk → memory → CPU) and optimize that first.
- Use batching to reduce overhead and context switching. For example, aggregate database writes rather than doing them one at a time.
- Separate the control plane from the data plane. For “hot loops” or performance‐critical code, create small, stateless helper functions accepting primitive types. This helps both the compiler and human reviewers spot redundant computations.

---

## Developer Experience

Our goal is to make our code easy to read, maintain, and extend.

### Naming Conventions

- **Go idioms:** For exported functions, use `CamelCase` (e.g., `ReadSector`), and for local functions and variables, use `camelCase`.
- Avoid abbreviations unless they are well-known acronyms (for example, use `HTTPStatus` rather than `httpStatus`).
- Include units and qualifiers in variable names at the end, so names like `latencyMsMax` (where “Ms” stands for milliseconds) are both clear and neatly aligned with related variables.
- When a function delegates work to helpers, consider prefixing the helper name with the parent function (e.g., `readSectorCallback`) to provide context.
- Place important constructs (like `main`) or critical types near the top of files for clarity.

### Comments and Commit Messages

- Write clear and complete comments. Comments should be full sentences with proper punctuation.
- Document why a decision was made, not just what the code does. Commit messages should convey context and rationale.

---

## Handling Errors

Errors must always be handled explicitly:

- In Go, this means checking returned errors immediately. Do not ignore them.
- Test error paths thoroughly, including non-fatal errors, to mitigate the risk of catastrophic failures found in distributed systems.
- Use assertions where appropriate to ensure that numbers or states remain within expected ranges.

---

## Avoiding Off-By-One Errors

Misunderstandings between zero-based indexes and one-based counts can be problematic:

- Clearly differentiate between an index, a count, and a size. For example, increment an index when calculating a count.
- Use helper functions for mathematical operations—consider naming them in a way that demonstrates intent (like `divExact`, `divFloor`, or `divCeil`).

---

## Formatting and Dependencies

- Run `go fmt` on all code. We also recommend using linters (such as `golangci-lint`) for consistency.
- Set a limit on line lengths (suggested maximum of 100 columns) to ensure clear visibility of code segments.
- Aim for a “zero dependencies” mindset where possible. Rely on Go’s standard library to reduce external risk and simplify builds.

---

## Tooling

- Choose tools that streamline development. When writing scripts or utilities, consider writing them in Go.
- A standardized toolchain reduces personal variations and ensures consistency across the team.

---

## Final Thoughts

In our adapted TigerStyle for Go, every detail—from strict assertions to deliberate naming and error handling—is designed to create robust, clear, and high-performance software. Remember: even though some of these practices (like assertions) aren’t common in Go, we include them as an essential part of our commitment to safety and clarity.

> "You don’t really suppose, do you, that all your adventures and escapes were managed by mere luck?"
> Keep iterating, keep improving, and let your code be both a precise tool and a piece of art.
